/*
  pygame - Python Game Library

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
*/

/*
 * Camera - webcam support for pygame
 * Author: Nirav Patel
 *
 * This module allows for use of v4l2 webcams in pygame.  The code is written
 * such that adding support for v4l or vfw cameras should be possible without
 * much modification of existing functions.  v4l2 functions are kept seperate
 * from functions available to pygame users and generic functions like
 * colorspace conversion.
 *
 * There is currently support for cameras that support MMAP and use pixelformats
 * of RGB24, RGB444, YUYV, SBGGR8, and YUV420.  To add support for additional
 * pixelformats, add them to v4l2_init_device and v4l2_process_image, and add
 * functions to convert the format to 24bit packed RGB, YUV, and HSV.
 */
 
#include "pygame.h"
#include "pygamedocs.h"
#include <structmember.h>
#include <stringobject.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <fcntl.h>              /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <asm/types.h>          /* for videodev2.h */

#include <linux/videodev2.h>

#define CLEAR(x) memset (&(x), 0, sizeof (x))
#define SAT(c) if (c & (~255)) { if (c < 0) c = 0; else c = 255; }
#define DEFAULT_WIDTH 640
#define DEFAULT_HEIGHT 480
#define RGB_OUT 1
#define YUV_OUT 2
#define HSV_OUT 4

struct buffer 
{
    void * start;
    size_t length;
};

typedef struct
{
    PyObject_HEAD
    char* device_name;
    unsigned long pixelformat;
    unsigned int color_out;
    struct buffer* buffers;
    unsigned int n_buffers;
    int width;
    int height;
    int size;
    int hflip;
    int vflip;
    int fd;
} PyCameraObject;

/* functions available to pygame users */
PyObject* surf_colorspace (PyObject* self, PyObject* arg);
PyObject* list_cameras (PyObject* self, PyObject* arg);
PyObject* camera_start (PyCameraObject* self);
PyObject* camera_stop (PyCameraObject* self);
PyObject* camera_get_controls (PyCameraObject* self);
PyObject* camera_set_controls (PyCameraObject* self, PyObject* arg);
PyObject* camera_get_size (PyCameraObject* self);
PyObject* camera_query_image (PyCameraObject* self);
PyObject* camera_get_image (PyCameraObject* self, PyObject* arg);
PyObject* camera_get_raw(PyCameraObject* self);
/* internal functions, potentially generic to input (v4l, v4l2, vfw, etc) */
static void colorspace (SDL_Surface *src, SDL_Surface *dst, int cspace);
static void rgb24_to_rgb24 (const void* src, void* dst, int length);
static void rgb444_to_rgb24 (const void* src, void* dst, int length);
static void rgb_to_yuv24 (const void* src, void* dst, int length, 
                          unsigned long source);
static void rgb_to_hsv24 (const void* src, void* dst, int length,
                          unsigned long source);
static void yuyv_to_rgb24 (const void* src, void* dst, int length);
static void yuyv_to_yuv24 (const void* src, void* dst, int length);
static void sbggr8_to_rgb24 (const void* src, void* dst, int width, int height);
static void yuv420_to_rgb24 (const void* src, void* dst, int width, int height);
static void yuv420_to_yuv24 (const void* src, void* dst, int width, int height);
/* internal functions specific to v4l2 */
static char** v4l2_list_cameras (int* num_devices);
static int v4l2_pixelformat (PyCameraObject* self, struct v4l2_format* fmt, 
                             unsigned long pixelformat);
static int v4l2_get_control (PyCameraObject* self, int id);
static int v4l2_set_control (PyCameraObject* self, int id, int value);
PyObject* v4l2_read_raw (PyCameraObject* self);
static int v4l2_xioctl (int fd, int request, void *arg);
static int v4l2_process_image (PyCameraObject* self, const void *image, 
                               unsigned int buffer_size, SDL_Surface* surf);
static int v4l2_query_buffer (PyCameraObject* self);
static int v4l2_read_frame (PyCameraObject* self, SDL_Surface* surf);
static int v4l2_stop_capturing (PyCameraObject* self);
static int v4l2_start_capturing (PyCameraObject* self);
static int v4l2_uninit_device (PyCameraObject* self);
static int v4l2_init_mmap (PyCameraObject* self);
static int v4l2_init_device (PyCameraObject* self);
static int v4l2_close_device (PyCameraObject* self);
static int v4l2_open_device (PyCameraObject* self);

/*
 * Functions available to pygame users.  The idea is to make these as simple as
 * possible, and merely have them call functions specific to the type of 
 * camera being used to do all the real work.  It currently just calls v4l2_*
 * functions, but it could check something like self->cameratype and depending
 * on the result, call v4l, v4l2, vfw, or other functions.
 */

/* colorspace() - Surface colorspace conversion */
PyObject* surf_colorspace (PyObject* self, PyObject* arg)
{
    PyObject *surfobj, *surfobj2;
    SDL_Surface* surf, *newsurf;
    char* color;
    int cspace;
    surfobj2 = NULL;

    /*get all the arguments*/
    if (!PyArg_ParseTuple (arg, "O!s|O!", &PySurface_Type, &surfobj, 
                           &color, &PySurface_Type, &surfobj2))
        return NULL;

    if (!strcmp(color, "YUV")) {
        cspace = YUV_OUT;
    } else if (!strcmp(color, "HSV")) {
        cspace = HSV_OUT;
    } else {
        return RAISE (PyExc_ValueError, "Incorrect colorspace value");
    }

    surf = PySurface_AsSurface (surfobj);
	
    if (!surfobj2)
    {
        newsurf = SDL_CreateRGBSurface (0, surf->w, surf->h, 24, 0xFF<<16, 
                                 0xFF<<8, 0xFF, 0);
        if (!newsurf)
            return NULL;
    }
    else
        newsurf = PySurface_AsSurface (surfobj2);

    /* check to see if the size is the same. */
    if (newsurf->w != surf->w || newsurf->h != surf->h)
        return RAISE (PyExc_ValueError, 
                      "Surfaces not the same width and height.");

    /* check to see if the format of the surface is the same. */
    if (surf->format->BitsPerPixel != 24 && newsurf->format->BitsPerPixel != 24)
        return RAISE (PyExc_ValueError, 
                      "Source and destination surfaces need to be 24 bit");

    SDL_LockSurface (newsurf);
    PySurface_Lock (surfobj);

    Py_BEGIN_ALLOW_THREADS;
    colorspace (surf, newsurf, cspace);
    Py_END_ALLOW_THREADS;

    PySurface_Unlock (surfobj);
    SDL_UnlockSurface (newsurf);

    if (surfobj2)
    {
        Py_INCREF (surfobj2);
        return surfobj2;
    }
    else
        return PySurface_New (newsurf);
}

/* list_cameras() - lists cameras available on the computer */
PyObject* list_cameras (PyObject* self, PyObject* arg)
{
    PyObject* ret_list;
    PyObject* string;
    char** devices;
    int num_devices, i;
    
    num_devices = 0;
    ret_list = NULL;
    ret_list = PyList_New (0);
    if (!ret_list)
        return NULL;

    devices = v4l2_list_cameras(&num_devices);
    
    for(i = 0; i < num_devices; i++) {
        string = PyString_FromString(devices[i]);
        PyList_Append(ret_list, string);
        Py_DECREF(string);
        free(devices[i]);
    }
    free(devices);
    
    return ret_list;
}

/* start() - opens, inits, and starts capturing on the camera */
PyObject* camera_start (PyCameraObject* self)
{
    if (v4l2_open_device(self) == 0)
        return NULL;
    if (v4l2_init_device(self) == 0)
        return NULL;
    if (v4l2_start_capturing(self) == 0)
        return NULL;
    Py_RETURN_NONE;
}

/* stop() - stops capturing, uninits, and closes the camera */
PyObject* camera_stop (PyCameraObject* self)
{
    if (v4l2_stop_capturing(self) == 0)
        return NULL;
    if (v4l2_uninit_device(self) == 0)
        return NULL;
    if (v4l2_close_device(self) == 0)
        return NULL;
    Py_RETURN_NONE;
}

/* get_controls() - gets current values of user controls */
/* TODO: Support brightness, contrast, and other common controls */
PyObject* camera_get_controls (PyCameraObject* self)
{
    int h, v;
    
    h = v4l2_get_control(self, V4L2_CID_HFLIP);
    if (h != -1)
        self->hflip = h;
    
    v = v4l2_get_control(self, V4L2_CID_VFLIP);
    if (v != -1)
        self->vflip = v;
    
    return Py_BuildValue ("(OO)", PyBool_FromLong(self->hflip), PyBool_FromLong(self->vflip));
}

/* set_controls() - changes camera settings if supported by the camera */
PyObject* camera_set_controls (PyCameraObject* self, PyObject* arg)
{
    int hflip, vflip;
    
    if (!PyArg_ParseTuple(arg, "|ii", &hflip, &vflip))
        return NULL;
        
    if (v4l2_set_control(self, V4L2_CID_HFLIP, hflip))
        self->hflip = hflip;
        
    if (v4l2_set_control(self, V4L2_CID_VFLIP, vflip))
        self->vflip = vflip;
    
    return Py_BuildValue ("(OO)", PyBool_FromLong(self->hflip), PyBool_FromLong(self->vflip));
}

/* get_size() - returns the dimensions of the images being recorded */
PyObject* camera_get_size (PyCameraObject* self)
{
    return Py_BuildValue ("(ii)", self->width, self->height);
}

/* query_image() - checks if a frame is ready */
PyObject* camera_query_image (PyCameraObject* self)
{
    return PyBool_FromLong(v4l2_query_buffer(self));
}

/* get_image() - returns an RGB Surface */
/* code to reuse Surface from René Dudfield */
PyObject* camera_get_image (PyCameraObject* self, PyObject* arg)
{
    SDL_Surface* surf = NULL;
    PyObject *surfobj = NULL;
    
    if (!PyArg_ParseTuple (arg, "|O!", &PySurface_Type, &surfobj))
        return NULL;

    if (!surfobj) {
        surf = SDL_CreateRGBSurface (0, self->width, self->height, 24, 0xFF<<16, 
                                 0xFF<<8, 0xFF, 0);
    } else {
        surf = PySurface_AsSurface (surfobj);
    }
    
    if (!surf)
        return NULL;
        
    if (surf->w != self->width || surf->h != self->height) {
        printf("width:%d:  height:%d:  surf->w:%d: surf->h:%d:\n", self->width, self->height, surf->w, surf->h);
        return RAISE (PyExc_ValueError, 
                      "Destination surface not the correct width or height.");
    }

    if (surf->format && surf->format->BitsPerPixel != 24) {
        return RAISE (PyExc_ValueError, 
                      "Destination surface not the correct bit depth.");
    }
    
    if (!v4l2_read_frame(self, surf))
        return NULL;
    
    if (!surf)
        return NULL;
        
    if (surfobj) {
        Py_INCREF (surfobj);
        return surfobj;
    } else {
        return PySurface_New (surf);
    }
}

/* get_raw() - returns an unmodified image as a string from the buffer */
PyObject* camera_get_raw(PyCameraObject* self)
{
    return v4l2_read_raw(self);
}

/*
 * Pixelformat conversion functions
 */
 
/* converts from rgb Surface to yuv or hsv */
/* TODO: Allow for conversion from yuv and hsv to all */
static void colorspace (SDL_Surface *src, SDL_Surface *dst, int cspace)
{   
    switch (cspace) {
        case YUV_OUT:
            rgb_to_yuv24 (src->pixels, dst->pixels, src->h * src->w, 0);
            break;
        case HSV_OUT:
            rgb_to_hsv24 (src->pixels, dst->pixels, src->h * src->w, 0);            
            break;
    }
}

/* converts pretty directly if its already RGB24 */
static void rgb24_to_rgb24 (const void* src, void* dst, int length)
{
    unsigned char *s, *d;

    s = (unsigned char *) src;
    d = (unsigned char *) dst;
    
    while (length--) {
        *d++ = *(s+2); /* blue */
        *d++ = *(s+1); /* green */
        *d++ = *s; /* red */
        s += 3;
    }
}

/* converts packed rgb to packed hsv. formulas modified from wikipedia */
static void rgb_to_hsv24 (const void* src, void* dst, int length, 
                          unsigned long source)
{
    unsigned char *s, *d;
    int r, g, b, max, min, delta;
    unsigned char p1, p2;

    s = (unsigned char *) src;
    d = (unsigned char *) dst;
    
    /* you could stick the if statement inside the loop, but I'm sacrificing a
       a few hundred bytes for a little performance */
    if (source == V4L2_PIX_FMT_RGB444) {
        while (length--) {
            p1 = *s++;
            p2 = *s++;
            b = p2 << 4;
            g = p1 & 0xF0;
            r = p1 << 4;
            max = MAX(MAX(r, g), b);
            min = MIN(MIN(r, g), b);
            delta = max - min;
            *d++ = max; /* value (similar to luminosity) */
            if (!delta) { /* grey, zero hue and saturation */
                *d++ = 0;
                *d++ = 0;
            } else {
                *d++ = 255*delta/max; /* saturation */
                if (r == max) {  /* set hue based on max color */
                    *d++ = 43*(g - b)/delta;
                } else if (g == max) {
                    *d++ = 85 + 43*(b - r)/delta;
                } else {
                    *d++ = 170 + 43*(r - g)/delta;
                }
            }
        }
    } else if (source == V4L2_PIX_FMT_RGB24) {
        while (length--) {
            r = *s++;
            g = *s++;
            b = *s++;
            max = MAX(MAX(r, g), b);
            min = MIN(MIN(r, g), b);
            delta = max - min;
            *d++ = max;
            if (!delta) {
                *d++ = 0;
                *d++ = 0;
            } else {
                *d++ = 255*delta/max;
                if (r == max) {
                    *d++ = 43*(g - b)/delta;
                } else if (g == max) {
                    *d++ = 85 + 43*(b - r)/delta;
                } else {
                    *d++ = 170 + 43*(r - g)/delta;
                }
            }
        }
    } else { /* for use as stage 2 in yuv or bayer to hsv, r and b switched */
        while (length--) {
            b = *s++;
            g = *s++;
            r = *s++;
            max = MAX(MAX(r, g), b);
            min = MIN(MIN(r, g), b);
            delta = max - min;
            *d++ = max;
            if (!delta) {
                *d++ = 0;
                *d++ = 0;
            } else {
                *d++ = 255*delta/max;
                if (r == max) {
                    *d++ = 43*(g - b)/delta;
                } else if (g == max) {
                    *d++ = 85 + 43*(b - r)/delta;
                } else {
                    *d++ = 170 + 43*(r - g)/delta;
                }
            }
        }
    }
}

/* convert packed rgb to yuv. Note that unlike many implementations of YUV,
   this has a full range of 0-255 for Y, not 16-235. Formulas from wikipedia */
static void rgb_to_yuv24 (const void* src, void* dst, int length, 
                          unsigned long source)
{
    unsigned char *s, *d;
    unsigned int r, g, b;
    unsigned char p1, p2;

    s = (unsigned char *) src;
    d = (unsigned char *) dst;
    
    if (source == V4L2_PIX_FMT_RGB444) {    
        while (length--) {
            p1 = *s++;
            p2 = *s++;
            b = p2 << 4;
            g = p1 & 0xF0;
            r = p1 << 4;
            *d++ = ((112*r-94*g-18*b + 128) >> 8) + 128;  /* V */
            *d++ = ((-38*r-74*g+112*b + 128) >> 8) + 128; /* U */
            *d++ = (77*r+150*g+29*b + 128) >> 8; /* Y */
        }
    } else if (source == V4L2_PIX_FMT_RGB24) {
        while (length--) {
            r = *s++;
            g = *s++;
            b = *s++;
            *d++ = ((112*r-94*g-18*b + 128) >> 8) + 128;
            *d++ = ((-38*r-74*g+112*b + 128) >> 8) + 128;
            *d++ = (77*r+150*g+29*b + 128) >> 8;
        }
    } else { /* for use as stage 2 in bayer to yuv, r and b switched */
        while (length--) {
            b = *s++;
            g = *s++;
            r = *s++;
            *d++ = ((112*r-94*g-18*b + 128) >> 8) + 128;
            *d++ = ((-38*r-74*g+112*b + 128) >> 8) + 128;
            *d++ = (77*r+150*g+29*b + 128) >> 8;
        }
    }
}

/* Converts from rgb444 (R444) to rgb24 (RGB3) */
static void rgb444_to_rgb24 (const void* src, void* dst, int length)
{
    unsigned char *s, *d;
    unsigned char p1, p2;
    
    s = (unsigned char *) src;
    d = (unsigned char *) dst;
    
    while (length--) {
        p1 = *s++;
        p2 = *s++;
        *d++ = p2 << 4; /* blue */
        *d++ = p1 & 0xF0; /* green */
        *d++ = p1 << 4; /* red */
    }
}   

/* convert from 4:2:2 YUYV interlaced to RGB24 */
/* modified from OpenCV, based on ccvt_yuyv_bgr32() from camstream */
static void yuyv_to_rgb24 (const void* src, void* dst, int length)
{
    unsigned char *s, *d;
    int i;
    int r, g, b, cr, cg, cb, y1, y2;

    i = length >> 1;
    s = (unsigned char *) src;
    d = (unsigned char *) dst;
    while (i--) {
        y1 = *s++;
        cb = ((*s - 128) * 454) >> 8;
        cg = (*s++ - 128) * 88;
        y2 = *s++;
        cr = ((*s - 128) * 359) >> 8;
        cg = (cg + (*s++ - 128) * 183) >> 8;

        r = y1 + cr;
        b = y1 + cb;
        g = y1 - cg;
        SAT(r);
        SAT(g);
        SAT(b);

        *d++ = b;
        *d++ = g;
        *d++ = r;

        r = y2 + cr;
        b = y2 + cb;
        g = y2 - cg;
        SAT(r);
        SAT(g);
        SAT(b);

        *d++ = b;
        *d++ = g;
        *d++ = r;
    }
}

/* turn yuyv into packed yuv. */
static void yuyv_to_yuv24 (const void* src, void* dst, int length)
{
    unsigned char *s, *d;
    int i;

    i = length >> 1;
    s = (unsigned char *) src;
    d = (unsigned char *) dst;
    while (i--) {
        *d++ = *(s+3); /* v */
        *d++ = *(s+1); /* u */
        *d++ = *s; /* y1 */
        *d++ = *(s+3); /* v */
        *d++ = *(s+1); /* u */
        *d++ = *(s+2); /* y2 */
        s += 4;
    }
}

/* Converts from 8 bit Bayer (BA81) to rgb24 (RGB3) */
/* from Sonix SN9C10x routines by Takafumi Mizuno <taka-qce@ls-a.jp> */
/* FIXME: Seems to be grayscale and kind of dark on the OLPC XO */
/*        Maybe the result of a different Bayer color order on the screen? */
/* TODO: Certainly not the most efficient way of doing this conversion. */
static void sbggr8_to_rgb24 (const void* src, void* dst, int width, int height)
{
    unsigned char *rawpt, *scanpt;
    int i = width * height;
    rawpt = (unsigned char*) src;
    scanpt = (unsigned char*) dst;

    while (i--) {
        if ( (i/width) % 2 == 0 ) {
            /* even row (BGBGBGBG)*/
            if ( (i % 2) == 0 ) {
                /* B */
                if ( (i > width) && ((i % width) > 0) ) {
                    *scanpt++ = *rawpt;                    /* B */
                    *scanpt++ = (*(rawpt-1)+*(rawpt+1)+
                    *(rawpt+width)+*(rawpt-width))/4;      /* G */
                    *scanpt++ = (*(rawpt-width-1)+*(rawpt-width+1)+
                    *(rawpt+width-1)+*(rawpt+width+1))/4;  /* R */
                } else {
                    /* first line or left column */
                    *scanpt++ = *rawpt;                             /* B */
                    *scanpt++ = (*(rawpt+1)+*(rawpt+width))/2;      /* G */
                    *scanpt++ = *(rawpt+width+1);                   /* R */
                }
            } else {
                /* (B)G */
                if ( (i > width) && ((i % width) < (width-1)) ) {
                    *scanpt++ = (*(rawpt-1)+*(rawpt+1))/2;          /* B */
                    *scanpt++ = *rawpt;                             /* G */
                    *scanpt++ = (*(rawpt+width)+*(rawpt-width))/2;  /* R */
                } else {
                    /* first line or right column */
                    *scanpt++ = *(rawpt-1);         /* B */
                    *scanpt++ = *rawpt;             /* G */
                    *scanpt++ = *(rawpt+width);     /* R */
                }
            }
        } else {
            /* odd row (GRGRGRGR) */
            if ( (i % 2) == 0 ) {
                /* G(R) */
                if ( (i < (width*(height-1))) && ((i % width) > 0) ) {
                    *scanpt++ = (*(rawpt+width)+*(rawpt-width))/2;  /* B */
                    *scanpt++ = *rawpt;                             /* G */
                    *scanpt++ = (*(rawpt-1)+*(rawpt+1))/2;          /* R */
                } else {
                    /* bottom line or left column */
                    *scanpt++ = *(rawpt-width);     /* B */
                    *scanpt++ = *rawpt;             /* G */
                    *scanpt++ = *(rawpt+1);         /* R */
                }
            } else {
                /* R */
                if ( i < (width*(height-1)) && ((i % width) < (width-1)) ) {
                    *scanpt++ = (*(rawpt-width-1)+*(rawpt-width+1)+                    
                    *(rawpt+width-1)+*(rawpt+width+1))/4;  /* B */
                    *scanpt++ = (*(rawpt-1)+*(rawpt+1)+
                    *(rawpt-width)+*(rawpt+width))/4;      /* G */
                    *scanpt++ = *rawpt;                    /* R */
                } else {
                    /* bottom line or right column */
                    *scanpt++ = *(rawpt-width-1);                   /* B */
                    *scanpt++ = (*(rawpt-1)+*(rawpt-width))/2;      /* G */
                    *scanpt++ = *rawpt;                             /* R */
                }
            }
        }
        rawpt++;
    }
}


/* convert from YUV 4:2:0 (YU12) to RGB24 */
/* modified from ccvt_c2.c from camstream by Tony Hague */
static void yuv420_to_rgb24 (const void* src, void* dst, int width, int height)
{
    const unsigned char *y1, *y2, *u, *v;
    unsigned char *l1, *l2;
    int r, g, b, cr, cg, cb, yp, j, i;
    
    l1 = (unsigned char *) dst;
    l2 = l1 + width * 3;
    y1 = (unsigned char *) src;
    y2 = y1 + width;
    u = y1 + width * height;
    v = u + (width * height) / 4;
    
    j = height / 2;
    while (j--) {
        i = width / 2;
        while (i--) {
            /* Since U & V are valid for 4 pixels, repeat code 4
               times for different Y */
            cb = ((*u-128) * 454)>>8;
            cr = ((*v-128) * 359)>>8;
            cg = ((*v-128) * 183 + (*u-128) * 88)>>8;

            yp = *y1++;
            r = yp + cr;
            b = yp + cb;
            g = yp - cg;
            SAT(r);
            SAT(g);
            SAT(b);
            *l1++ = b;
            *l1++ = g;
            *l1++ = r;
            
            yp = *y1++;
            r = yp + cr;
            b = yp + cb;
            g = yp - cg;
            SAT(r);
            SAT(g);
            SAT(b);
            *l1++ = b;
            *l1++ = g;
            *l1++ = r;

            yp = *y2++;
            r = yp + cr;
            b = yp + cb;
            g = yp - cg;
            SAT(r);
            SAT(g);
            SAT(b);
            *l2++ = b;
            *l2++ = g;
            *l2++ = r;
            
            yp = *y2++;
            r = yp + cr;
            b = yp + cb;
            g = yp - cg;
            SAT(r);
            SAT(g);
            SAT(b);
            *l2++ = b;
            *l2++ = g;
            *l2++ = r;
            
            u++;
            v++;
        }
        y1 = y2;
        y2 += width;
        l1 = l2;
        l2 += width*3;
    }
}

/* turn yuv420 into packed yuv. */
static void yuv420_to_yuv24 (const void* src, void* dst, int width, int height)
{
    const unsigned char *y1, *y2, *u, *v;
    unsigned char *l1, *l2;
    int j, i;
    
    l1 = (unsigned char *) dst;
    l2 = l1 + width * 3;
    y1 = (unsigned char *) src;
    y2 = y1 + width;
    u = y1 + width * height;
    v = u + (width * height) / 4;
    
    j = height / 2;
    while (j--) {
        i = width / 2;
        while (i--) {
            *l1++ = *v;
            *l1++ = *u;
            *l1++ = *y1++;
            *l1++ = *v;
            *l1++ = *u;
            *l1++ = *y1++;
            *l2++ = *v;
            *l2++ = *u;
            *l2++ = *y2++;
            *l2++ = *v++;
            *l2++ = *u++;
            *l2++ = *y2++;
        }
        y1 = y2;
        y2 += width;
        l1 = l2;
        l2 += width*3;
    }
}

static char** v4l2_list_cameras (int* num_devices)
{
    char** devices;
    char* device;
    int num, i, fd;
    
    num = *num_devices;
    
    devices = (char**) malloc(sizeof(char *)*65);
    
    device = (char*) malloc(sizeof(char)*13);
    strcpy(device,"/dev/video");
    fd = open(device, O_RDONLY);
    if (fd != -1) {
        devices[num] = device;
        num++;
        device = (char*) malloc(sizeof(char)*13);
    }
    close(fd);
    /* v4l2 cameras can be /dev/video and /dev/video0 to /dev/video63 */
    for (i = 0; i < 64; i++) {
        sprintf(device,"/dev/video%d",i);
        fd = open(device, O_RDONLY);
        if (fd != -1) {
            devices[num] = device;
            num++;
            device = (char*) malloc(sizeof(char)*13);
        }
        close(fd);
    }
    
    if (num == *num_devices) {
        free(device);
    } else {
        *num_devices = num;
    }
    
    return devices;
}

/* A wrapper around a VIDIOC_S_FMT ioctl to check for format compatibility */
static int v4l2_pixelformat (PyCameraObject* self, struct v4l2_format* fmt, 
                             unsigned long pixelformat)
{
    fmt->fmt.pix.pixelformat = pixelformat;

    if (-1 == v4l2_xioctl (self->fd, VIDIOC_S_FMT, fmt)) {
        return 0;
    }
    
    if (fmt->fmt.pix.pixelformat == pixelformat) {
    	return 1;
    } else {
        return 0;
    }
}

/* gets the value of a specific camera control if available */
/* FIXME: returning -1 is bad. controls like hue will return -1 legitimately */
static int v4l2_get_control (PyCameraObject* self, int id)
{
    struct v4l2_control control;
    CLEAR(control);
    
    control.id = id;
    
    if (-1 == v4l2_xioctl (self->fd, VIDIOC_G_CTRL, &control)) {
        return -1;
    }
    
    return control.value;
}

/* sets a control if supported. the camera may round the value */
static int v4l2_set_control (PyCameraObject* self, int id, int value)
{
    struct v4l2_control control;
    CLEAR(control);
    
    control.id = id;
    control.value = value;
    
    if (-1 == v4l2_xioctl (self->fd, VIDIOC_S_CTRL, &control)) {
        return 0;
    }
    
    return 1;
}

/* returns a string of the buffer from the camera */
/* TODO: fold this into the regular read_frame. lots of duplicate code */
PyObject* v4l2_read_raw (PyCameraObject* self)
{
    struct v4l2_buffer buf;
    PyObject* raw;

    CLEAR (buf);

    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;

    if (-1 == v4l2_xioctl (self->fd, VIDIOC_DQBUF, &buf)) {
        PyErr_Format(PyExc_SystemError, "ioctl(VIDIOC_DQBUF) failure : %d, %s",
                     errno, strerror (errno));
        return NULL;
    }

    assert (buf.index < self->n_buffers);

    raw = PyString_FromStringAndSize(self->buffers[buf.index].start, 
                                     self->buffers[buf.index].length);

    if (-1 == v4l2_xioctl (self->fd, VIDIOC_QBUF, &buf)) {
        PyErr_Format(PyExc_SystemError, "ioctl(VIDIOC_QBUF) failure : %d, %s",
                     errno, strerror (errno));
        return NULL;
    }
    return raw;
}

/*
 * Functions for v4l2 cameras.
 * This code is based partly on pyvideograb by Laurent Pointal at
 * http://laurent.pointal.org/python/projets/pyvideograb
 * the v4l2 capture example at
 * http://www.linuxtv.org/download/video4linux/API/V4L2_API/spec/
 * and the HighGUI library in OpenCV.
 */

static int v4l2_xioctl (int fd, int request, void *arg)
{
    int r;

    do r = ioctl (fd, request, arg);
    while (-1 == r && EINTR == errno);
    
    return r;
}

/* sends the image to the conversion function based on input format and 
   desired output format.  Note that some of the less common conversions are
   currently two step processes. */
/* TODO: Write single step conversions where they may actually be useful */
static int v4l2_process_image (PyCameraObject* self, const void *image, 
                               unsigned int buffer_size, SDL_Surface* surf)
{

    if (!surf)
        return 0;
        
    SDL_LockSurface (surf);
    
    switch (self->pixelformat) {
        case V4L2_PIX_FMT_RGB24:
            if (buffer_size >= self->size * 3) {
                switch (self->color_out) {
                    case RGB_OUT:
                        rgb24_to_rgb24(image, surf->pixels, self->size);
                        break;
                    case HSV_OUT:
                        rgb_to_hsv24(image, surf->pixels, self->size, V4L2_PIX_FMT_RGB24);
                        break;
                    case YUV_OUT:
                        rgb_to_yuv24(image, surf->pixels, self->size, V4L2_PIX_FMT_RGB24);
                        break;
                }
            } else {
                SDL_UnlockSurface (surf);
                return 0;
            }
            break;
        case V4L2_PIX_FMT_RGB444:
            if (buffer_size >= self->size * 2) {
                switch (self->color_out) {
                    case RGB_OUT:
                        rgb444_to_rgb24(image, surf->pixels, self->size);
                        break;
                    case HSV_OUT:
                        rgb_to_hsv24(image, surf->pixels, self->size, V4L2_PIX_FMT_RGB444);
                        break;
                    case YUV_OUT:
                        rgb_to_yuv24(image, surf->pixels, self->size, V4L2_PIX_FMT_RGB444);
                        break;
                }
            } else {
                SDL_UnlockSurface (surf);
                return 0;
            }
            break;
        case V4L2_PIX_FMT_YUYV:
            if (buffer_size >= self->size * 2) {
                switch (self->color_out) {
                    case YUV_OUT:
                        yuyv_to_yuv24(image, surf->pixels, self->size);
                        break;
                    case RGB_OUT:
                        yuyv_to_rgb24(image, surf->pixels, self->size);
                        break;
                    case HSV_OUT:
                        yuyv_to_rgb24(image, surf->pixels, self->size);
                        rgb_to_hsv24(surf->pixels, surf->pixels, self->size, V4L2_PIX_FMT_YUYV);
                        break;
                }
            } else {
                SDL_UnlockSurface (surf);
                return 0;
            }
            break;
        case V4L2_PIX_FMT_SBGGR8:
            if (buffer_size >= self->size) {
                switch (self->color_out) {
                    case RGB_OUT:
                        sbggr8_to_rgb24(image, surf->pixels, self->width, self->height);
                        break;
                    case HSV_OUT:
                        sbggr8_to_rgb24(image, surf->pixels, self->width, self->height);
                        rgb_to_hsv24(surf->pixels, surf->pixels, self->size, V4L2_PIX_FMT_SBGGR8);
                        break;
                    case YUV_OUT:
                        sbggr8_to_rgb24(image, surf->pixels, self->width, self->height);
                        rgb_to_yuv24(surf->pixels, surf->pixels, self->size, V4L2_PIX_FMT_SBGGR8);
                        break;
                }
            } else {
                SDL_UnlockSurface (surf);
                return 0;
            }
            break;
        case V4L2_PIX_FMT_YUV420:
            if (buffer_size >= (self->size * 3) / 2) {
                switch (self->color_out) {
                    case YUV_OUT:
                        yuv420_to_yuv24(image, surf->pixels, self->width, self->height);
                        break;
                    case RGB_OUT:
                        yuv420_to_rgb24(image, surf->pixels, self->width, self->height);
                        break;
                    case HSV_OUT:
                        yuv420_to_rgb24(image, surf->pixels, self->width, self->height);
                        rgb_to_hsv24(surf->pixels, surf->pixels, self->size, V4L2_PIX_FMT_YUV420);
                        break;
                }
            } else {
                SDL_UnlockSurface (surf);
                return 0;
            }
            break;
    }
    SDL_UnlockSurface (surf);
    return 1;
}

/* query each buffer to see if it contains a frame ready to take */
/* FIXME: There needs to be a better way to implement non-blocking frame
   grabbing than only doing a get_image if query_image returns true. Many
   cameras will always return false, and will only respond to blocking calls. */
static int v4l2_query_buffer (PyCameraObject* self)
{
    int i;
    
    for (i = 0; i < self->n_buffers; ++i) {
        struct v4l2_buffer buf;

        CLEAR (buf);

        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index = i;

        if (-1 == v4l2_xioctl (self->fd, VIDIOC_QUERYBUF, &buf)) {
            PyErr_Format(PyExc_MemoryError, "ioctl(VIDIOC_QUERYBUF) failure : %d, %s",
                errno, strerror (errno));
            return 0;
        }

        /*  is there a buffer on outgoing queue ready for us to take? */
        if (buf.flags & V4L2_BUF_FLAG_DONE)
            return 1;
    }
    
    /* no buffer ready to take */
    return 0;
}

static int v4l2_read_frame (PyCameraObject* self, SDL_Surface* surf)
{
    struct v4l2_buffer buf;

    CLEAR (buf);

    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;

    if (-1 == v4l2_xioctl (self->fd, VIDIOC_DQBUF, &buf)) {
        PyErr_Format(PyExc_SystemError, "ioctl(VIDIOC_DQBUF) failure : %d, %s",
                     errno, strerror (errno));
        return 0;
    }

    assert (buf.index < self->n_buffers);

    if (!v4l2_process_image (self, self->buffers[buf.index].start, 
                             self->buffers[buf.index].length, surf)) {
        PyErr_Format(PyExc_SystemError, "image processing error");
        return 0;
    }

    if (-1 == v4l2_xioctl (self->fd, VIDIOC_QBUF, &buf)) {
        PyErr_Format(PyExc_SystemError, "ioctl(VIDIOC_QBUF) failure : %d, %s",
                     errno, strerror (errno));
        return 0;
    }
    return 1;
}

static int v4l2_stop_capturing (PyCameraObject* self)
{
    enum v4l2_buf_type type;

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (-1 == v4l2_xioctl (self->fd, VIDIOC_STREAMOFF, &type)) {
        PyErr_Format(PyExc_SystemError, "ioctl(VIDIOC_STREAMOFF) failure : %d, %s",
                     errno, strerror (errno));
        return 0;
    }

    return 1;
}

static int v4l2_start_capturing (PyCameraObject* self)
{
    unsigned int i;
    enum v4l2_buf_type type;

    for (i = 0; i < self->n_buffers; ++i) {
        struct v4l2_buffer buf;

        CLEAR (buf);

        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index = i;

        if (-1 == v4l2_xioctl (self->fd, VIDIOC_QBUF, &buf)) {
            PyErr_Format(PyExc_EnvironmentError, "ioctl(VIDIOC_QBUF) failure : %d, %s",
                         errno, strerror (errno));
            return 0;
        }
    }

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (-1 == v4l2_xioctl (self->fd, VIDIOC_STREAMON, &type)) {
        PyErr_Format(PyExc_EnvironmentError, "ioctl(VIDIOC_STREAMON) failure : %d, %s",
                      errno, strerror (errno));
        return 0;
    }

    return 1;
}

static int v4l2_uninit_device (PyCameraObject* self)
{
    unsigned int i;

    for (i = 0; i < self->n_buffers; ++i) {
        if (-1 == munmap (self->buffers[i].start, self->buffers[i].length)) {
            PyErr_Format(PyExc_MemoryError, "munmap failure: %d, %s",
                         errno, strerror (errno));
            return 0;
        }
    }

    free (self->buffers);

    return 1;
}

static int v4l2_init_mmap (PyCameraObject* self)
{
    struct v4l2_requestbuffers req;
    
    CLEAR (req);

    /* 2 is the minimum possible, and some drivers will force a higher count.
       It will likely result in buffer overruns, but for purposes of gaming,
       it is probably better to drop frames than get old frames. */
    req.count = 2;
    req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory = V4L2_MEMORY_MMAP;

    if (-1 == v4l2_xioctl (self->fd, VIDIOC_REQBUFS, &req)) {
        if (EINVAL == errno) {
            PyErr_Format(PyExc_MemoryError, "%s does not support memory mapping",
                self->device_name);
            return 0;
        }
        else {
            PyErr_Format(PyExc_MemoryError, "ioctl(VIDIOC_REQBUFS) failure : %d, %s",
                errno, strerror (errno));
            return 0;
        }
    }

    if (req.count < 2) {
        PyErr_Format(PyExc_MemoryError, "Insufficient buffer memory on %s\n",
            self->device_name);
        return 0;
    }

    self->buffers = calloc (req.count, sizeof (*self->buffers));

    if (!self->buffers) {
        PyErr_Format(PyExc_MemoryError, "Out of memory");
        return 0;
    }

    for (self->n_buffers = 0; self->n_buffers < req.count; ++self->n_buffers) {
        struct v4l2_buffer buf;

        CLEAR (buf);

        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index = self->n_buffers;

        if (-1 == v4l2_xioctl (self->fd, VIDIOC_QUERYBUF, &buf)) {
            PyErr_Format(PyExc_MemoryError, "ioctl(VIDIOC_QUERYBUF) failure : %d, %s",
                errno, strerror (errno));
            return 0;
        }

        self->buffers[self->n_buffers].length = buf.length;
        self->buffers[self->n_buffers].start =
            mmap (NULL /* start anywhere */,
                buf.length,
                PROT_READ | PROT_WRITE /* required */,
                MAP_SHARED /* recommended */,
                self->fd, buf.m.offset);

        if (MAP_FAILED == self->buffers[self->n_buffers].start) {
            PyErr_Format(PyExc_MemoryError, "mmap failure : %d, %s",
                errno, strerror (errno));
            return 0;
        }
    }

    return 1;
}

static int v4l2_init_device (PyCameraObject* self)
{
    struct v4l2_capability cap;
    struct v4l2_format fmt;
    unsigned int min;

    if (-1 == v4l2_xioctl (self->fd, VIDIOC_QUERYCAP, &cap)) {
        if (EINVAL == errno) {
            PyErr_Format(PyExc_SystemError, "%s is not a V4L2 device",
                self->device_name);
            return 0;
        }
        else {
            PyErr_Format(PyExc_SystemError, "ioctl(VIDIOC_QUERYCAP) failure : %d, %s",
                errno, strerror (errno));
            return 0;
        }
    }

    if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
        PyErr_Format(PyExc_SystemError, "%s is not a video capture device",
            self->device_name);
        return 0;
    }

    if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
        PyErr_Format(PyExc_SystemError, "%s does not support streaming i/o",
            self->device_name);
        return 0;
    }

    CLEAR (fmt);

    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fmt.fmt.pix.width = self->width;
    fmt.fmt.pix.height = self->height;
    fmt.fmt.pix.field = V4L2_FIELD_ANY;
    
    /* Find the pixelformat supported by the camera that will take the least
       processing power to convert to the desired output.  Thus, for YUV out,
       YUYVand YUV420 are first, while for RGB and HSV, the packed RGB formats
       are first. */
    switch (self->color_out) {
        case YUV_OUT:
            if (v4l2_pixelformat(self, &fmt, V4L2_PIX_FMT_YUYV)) {
                self->pixelformat = V4L2_PIX_FMT_YUYV;
            } else if (v4l2_pixelformat(self, &fmt, V4L2_PIX_FMT_YUV420)) {
                self->pixelformat = V4L2_PIX_FMT_YUV420;
            } else if (v4l2_pixelformat(self, &fmt, V4L2_PIX_FMT_RGB24)) {
                self->pixelformat = V4L2_PIX_FMT_RGB24;
            } else if (v4l2_pixelformat(self, &fmt, V4L2_PIX_FMT_RGB444)) {
                self->pixelformat = V4L2_PIX_FMT_RGB444;
            } else if (v4l2_pixelformat(self, &fmt, V4L2_PIX_FMT_SBGGR8)) {
                self->pixelformat = V4L2_PIX_FMT_SBGGR8;
            } else {
                PyErr_Format(PyExc_SystemError,
                           "ioctl(VIDIOC_S_FMT) failure: no supported formats");
                return 0;
            }
            break;
        default:
            if (v4l2_pixelformat(self, &fmt, V4L2_PIX_FMT_RGB24)) {
                self->pixelformat = V4L2_PIX_FMT_RGB24;
            } else if (v4l2_pixelformat(self, &fmt, V4L2_PIX_FMT_RGB444)) {
                self->pixelformat = V4L2_PIX_FMT_RGB444;
            } else if (v4l2_pixelformat(self, &fmt, V4L2_PIX_FMT_YUYV)) {
                self->pixelformat = V4L2_PIX_FMT_YUYV;
            } else if (v4l2_pixelformat(self, &fmt, V4L2_PIX_FMT_SBGGR8)) {
                self->pixelformat = V4L2_PIX_FMT_SBGGR8;
            } else if (v4l2_pixelformat(self, &fmt, V4L2_PIX_FMT_YUV420)) {
                self->pixelformat = V4L2_PIX_FMT_YUV420;
            } else {
                PyErr_Format(PyExc_SystemError,
                           "ioctl(VIDIOC_S_FMT) failure: no supported formats");
                return 0;
            }
            break;
    }    

    /* Note VIDIOC_S_FMT may change width and height. */
    self->width = fmt.fmt.pix.width;
    self->height = fmt.fmt.pix.height;
    self->size = self->width * self->height;
    self->pixelformat = fmt.fmt.pix.pixelformat;

    /* Buggy driver paranoia. */
    min = fmt.fmt.pix.width * 2;
    if (fmt.fmt.pix.bytesperline < min)
        fmt.fmt.pix.bytesperline = min;
    min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
    if (fmt.fmt.pix.sizeimage < min)
        fmt.fmt.pix.sizeimage = min;

    v4l2_init_mmap (self);

    return 1;
}

static int v4l2_close_device (PyCameraObject* self)
{
    if (self->fd==-1)
        return 1;

    if (-1 == close (self->fd)) {
        PyErr_Format(PyExc_SystemError, "Cannot close '%s': %d, %s",
            self->device_name, errno, strerror (errno));
        return 0;
    }
    self->fd = -1;

    return 1;
}

static int v4l2_open_device (PyCameraObject* self)
{
    struct stat st;

    if (-1 == stat (self->device_name, &st)) {
        PyErr_Format(PyExc_SystemError, "Cannot identify '%s': %d, %s",
            self->device_name, errno, strerror (errno));
        return 0;
    }

    if (!S_ISCHR (st.st_mode)) {
        PyErr_Format(PyExc_SystemError, "%s is no device",self->device_name);
        return 0;
    }

    self->fd = open (self->device_name, O_RDWR /* required | O_NONBLOCK */, 0);

    if (-1 == self->fd) {
        PyErr_Format(PyExc_SystemError, "Cannot open '%s': %d, %s",
            self->device_name, errno, strerror (errno));
        return 0;
    }

    return 1;
}

/*
 * Python API stuff
 */
 
static PyMethodDef cameraobj_builtins[] =
{
    { "start", (PyCFunction) camera_start, METH_NOARGS, DOC_CAMERASTART },
    { "stop", (PyCFunction) camera_stop, METH_NOARGS, DOC_CAMERASTOP },
    { "get_controls", (PyCFunction) camera_get_controls, METH_NOARGS, DOC_CAMERAGETCONTROLS },
    { "set_controls", (PyCFunction) camera_set_controls, METH_VARARGS, DOC_CAMERASETCONTROLS },
    { "get_size", (PyCFunction) camera_get_size, METH_NOARGS, DOC_CAMERAGETSIZE },
    { "query_image", (PyCFunction) camera_query_image, METH_NOARGS, DOC_CAMERAQUERYIMAGE },
    { "get_image", (PyCFunction) camera_get_image, METH_VARARGS, DOC_CAMERAGETIMAGE },
    { "get_raw", (PyCFunction) camera_get_raw, METH_NOARGS, DOC_CAMERAGETRAW },
    { NULL, NULL, 0, NULL }
};

static void camera_dealloc (PyObject* self)
{
    free(((PyCameraObject*) self)->device_name);
    PyObject_DEL (self);
}

static PyObject* camera_getattr(PyObject* self, char* attrname)
{
    return Py_FindMethod(cameraobj_builtins, self, attrname);
}

static PyTypeObject PyCamera_Type =
{
    PyObject_HEAD_INIT(NULL)
    0,
    "Camera",
    sizeof(PyCameraObject),
    0,
    camera_dealloc,
    0,
    camera_getattr,
    NULL,			/*setattr*/
    NULL,			/*compare*/
    NULL,			/*repr*/
    NULL,			/*as_number*/
    NULL,			/*as_sequence*/
    NULL,			/*as_mapping*/
    (hashfunc)NULL, 		/*hash*/
    (ternaryfunc)NULL,		/*call*/
    (reprfunc)NULL, 		/*str*/
    0L,0L,0L,0L,
    DOC_PYGAMECAMERACAMERA /* Documentation string */
};

static PyObject* Camera (PyCameraObject* self, PyObject* arg)
{
    int w, h;
    char* dev_name = NULL;
    char* color = NULL;
    PyCameraObject *cameraobj;
    
    w = DEFAULT_WIDTH;
    h = DEFAULT_HEIGHT;
    
    if (!PyArg_ParseTuple(arg, "s|(ii)s", &dev_name, &w, &h, &color))
        return NULL;
    
    cameraobj = PyObject_NEW (PyCameraObject, &PyCamera_Type);
    
    if (cameraobj) {
        cameraobj->device_name = (char*) malloc((strlen(dev_name)+1)*sizeof(char));
        strcpy(cameraobj->device_name, dev_name);
        cameraobj->pixelformat = 0;
        if (color) {
            if (!strcmp(color, "YUV")) {
                cameraobj->color_out = YUV_OUT;
            } else if (!strcmp(color, "HSV")) {
                cameraobj->color_out = HSV_OUT;
            } else {
                cameraobj->color_out = RGB_OUT;
            }
        } else {
            cameraobj->color_out = RGB_OUT;
        }
        cameraobj->buffers = NULL;
        cameraobj->n_buffers = 0;
        cameraobj->width = w;
        cameraobj->height = h;
        cameraobj->size = 0;
        cameraobj->hflip = 0;
        cameraobj->vflip = 0;
        cameraobj->fd = -1;
    }
    
    return (PyObject*)cameraobj;
}

static PyMethodDef camera_builtins[] =
{
    { "colorspace", surf_colorspace, METH_VARARGS, DOC_PYGAMECAMERACOLORSPACE },
    { "list_cameras", list_cameras, METH_NOARGS, DOC_PYGAMECAMERALISTCAMERAS },
    { "Camera", (PyCFunction) Camera, METH_VARARGS, DOC_PYGAMECAMERACAMERA },
    { NULL, NULL, 0, NULL }
};
 
void initcamera(void)
{
  PyObject *module, *dict;
  PyType_Init(PyCamera_Type);
  
  /* create the module */
  module = Py_InitModule3("camera", camera_builtins, DOC_PYGAMECAMERA);
  dict = PyModule_GetDict(module);
  PyDict_SetItemString(dict, "CameraType", (PyObject *)&PyCamera_Type);
  import_pygame_base ();
  import_pygame_surface ();
}
